import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;

import kernelization.MatchingAndComputationTimes;
import kernelization.MatchingMaker;

import org.jgrapht.UndirectedGraph;
import org.jgrapht.graph.DefaultEdge;

public class ComputationTimeB5 {

	// matching algorithm
	final private static int algo = 5;

	public static void main(String[] args) {
		
		// args
		
		if (args.length == 0) {
			System.out.println("Called program without arguments. Must at least specify the directory with input graph files.");
			System.out.println("For example: dataset/SNAP/test/");
			System.out.println("Input graphs must be in DIMACS format.");
			System.out.println("Set first argument to '-help' to list further options.");
			return;
		}
			
		if (args[0].equals("-help")) {
			System.out.println("The first argument must always be the directory with input files.");
			System.out.println("Following options can be added after the directory:");
			
			System.out.println("-kX   	     \t Kernelization method. Replace X by 0-3. Default is 1.");
			System.out.println("             \t\t 0: Without data reduction rules.");
			System.out.println("             \t\t 1: Data reduction rules for vertices");
			System.out.println("             \t\t    with degree 0, 1, and 2");
			System.out.println("             \t\t    without degree restriction.");
			System.out.println("             \t\t 2: Data reduction rules for vertices");
			System.out.println("             \t\t    with degree 0, 1, and 2");
			System.out.println("             \t\t    with degree restriction used for the merging");
			System.out.println("             \t\t    the approach of kernelization method 1.");
			System.out.println("             \t\t 3: Data reduction rules for vertices");
			System.out.println("             \t\t    with degree 0, 1, and 2");
			System.out.println("             \t\t    with degree restriction used for the merging");
			System.out.println("             \t\t    the special approach.");
			
			System.out.println("-outputfile  \t Output file for results. Default is 'Results_ComputationTime'.");
			System.out.println("-timelimit   \t Limit in milliseconds for the computation time of one graph.");
			System.out.println("             \t Default is one hour.");
			System.out.println("-abortfactor \t Computation will be finished if");
			System.out.println("             \t\t size of actual file > ( abort factor * size of first file");
			System.out.println("             \t\t whose computation time was over time limit ).");
			System.out.println("             \t Default is 1.");
			System.out.println("To show this usage information, set first argument to '-help'.");
			return;
		}
		
		
		// default values for args:
		
		// time limit in milliseconds for computation of one file
		int timeout = 1000 * 60 * 60;
		// abort factor: computation will be finished if 
		// size of actual file > ( abort factor * size of first file whose computation time was over time limit )
		double abortFactor = 1;
		// output file
		File outputFile = new File("Results_ComputationTime.log");
		// kernelization method
		int method = 1; 
		
		
		// obligatory argument:

		// directory with graphs
		File[] list = new File(args[0]).listFiles();	
		
		if (list == null) {
			System.out.println("Directory doesn't exist.");
			System.out.println("First argument must be the directory with input graph files.");
			System.out.println("For example: dataset/SNAP/test/");
			System.out.println("Input graphs must be in DIMACS format.");
			System.out.println("Set first argument to '-help' for usage information.");
			return;
		}

		if (list.length == 0) {
			System.out.println("Directory has no files.");
			return;
		}
		
		
		// optional arguments:
		
		if (args.length > 1) {
			boolean wellFormed = true;
			for (int arg = 1; arg < args.length; arg++) {
				try {
					switch (args[arg]) {
						case "-k0"	:		method = 0; break;
						case "-k1"	:		method = 1; break;
						case "-k2"	:		method = 2; break;
						case "-k3"	:		method = 3; break;
						
						case "-outputfile" : 
											if (arg+1 == args.length) wellFormed = false;
											else outputFile = new File(args[++arg] + ".log");
											break;
											
						case "-timelimit" :	if (arg+1 == args.length) wellFormed = false;
											else timeout = Integer.parseInt(args[++arg]);
											break;
											
						case "-abortfactor" :	if (arg+1 == args.length) wellFormed = false;
											else abortFactor = Double.parseDouble(args[++arg]);
											break;
					} 
				} catch (NumberFormatException e) {
					wellFormed = false; break;
				}
				if (!wellFormed) break;
			}
			if (!wellFormed) {
				System.out.println("Program does not understand arguments of program call.");
				System.out.println("Set first argument to '-help' for usage information.");
				return;
			}
		}	
		
		System.out.println("Computation is started with the following parameters:");
		if (method == 0)
			System.out.println("- without kernelization");
		else {
			System.out.println("- kernelization method " + method + ":\n\tdata reduction rules for vertices with degree 0, 1, and 2 ");
			switch (method) {
			case 1	:		System.out.println("\twithout degree restriction"); break;
			case 2	:		System.out.println("\twith degree restriction used for the merging the approach of kernelization method 1"); break;
			case 3	:		System.out.println("\twith degree restriction used for the merging the special approach"); break;
			}
		}
		System.out.println("- time limit: "
				+ (timeout / 1000) / 60 + " min. " 
				+ (timeout / 1000) % 60 + " sec. "
				+ timeout % 1000 + " ms. ");
		System.out.println("- abort factor: " + abortFactor);
		System.out.print("Please wait. "); 
		
		
		///////////////////////////////////////////////////////////////////////////////////

		try {
			// log-writer
			BufferedWriter writer = new BufferedWriter(new FileWriter(outputFile));
			// headline
			writer.write("graph" + "\t");
			writer.write("nodes" + "\t");
			writer.write("edges" + "\t");
			writer.write("matching size" + "\t");
			writer.write("DRR [s]" + "\t");
			writer.write("matching [s]" + "\t");
			writer.write("postprocessing [s]");
			if (method != 0)				
				writer.write("\t" + "total kernelizaion " + method + " [s]");
			writer.write("\n");

			// sort files from smallest to biggest
			Arrays.sort(list, new Comparator<File>() {
				@Override
				public int compare(File f1, File f2) {
					return Long.compare(f1.length(), f2.length());
				}
			});

			System.out.println("Working on:");
			double sizeOfFirstFileOverTimeLimit = -1;
			int counter = 0;

			// for every graph in data set
			for (File file : list) {
				if (sizeOfFirstFileOverTimeLimit != -1 && file.length() > abortFactor * sizeOfFirstFileOverTimeLimit){
					System.out.println("Computation is stopped because sizes of all remaining files in this directories are more than\n"
							+ " abort factor * size of first file whose computation time was over time limit."
							+ " (" + abortFactor + " * " + sizeOfFirstFileOverTimeLimit / 1024 + " KB )");
					break;
				}

				if (file.isFile() && ! file.isHidden())
				{
					counter++;

					// working on
					System.out.print("  " + counter + ") " + file.getName() + " (" + file.length() / 1024 + " KB), "
							+ "started " + DateFormat.getDateTimeInstance().format(new Date()));

					// stop after timeout
					Testing tests = new Testing();
					tests.setFile(file);
					tests.setKM(method);
					tests.start();
					try {
						tests.join(timeout);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
					tests.interrupt();		

					// if no exception
					if (tests.hasNoException()) {
						// if not stopped
						if (tests.isCompleted()) {
							System.out.println(", completed " + DateFormat.getDateTimeInstance().format(new Date()));
							// write graph and number of nodes and edges
							writer.write(file.getName() + "\t");
							writer.write(tests.getVertexNumber() + "\t");
							writer.write(tests.getEdgeNumber() + "\t");	

							// write size of matching
							writer.write(tests.getSize() + "\t");

							// write computation times
							double[] time = tests.getTime();				
							// write computation times of kernelizations in details
							if (method != 0) {			
								for (int i = 1; i <= 3; i++)
									writer.write(time[i] + "\t");
								// total
								writer.write(time[0] + "\n");
							}
							else
								writer.write("-\t" + time[2] + "\t-\n");
						}
						// if stopped because of timeout
						else {
							System.out.print(", ");
							System.err.println("stopped " + DateFormat.getDateTimeInstance().format(new Date()));
							if (sizeOfFirstFileOverTimeLimit == -1)
								sizeOfFirstFileOverTimeLimit = file.length();
						}
					}

					if (list[list.length - 1] == file)
						System.out.println("Computation is finished.");
				}
			}

			//close stream
			writer.close();
		} catch (IOException e) {
			System.err.println("IOException. Computation is breaked.");
			return;
		} catch (NullPointerException e) {
			System.err.println("NullPointerException. Computation is breaked.");
			return;
		}

		System.out.println("Results are in the file \"" + outputFile.getName() + "\".");
	}

	public static class Testing extends Thread {

		private File file;
		private int method = 1;
		private int vertexNumber = -1;
		private int edgeNumber = -1;
		private int size = -1;
		private double[] time = {-1, -1, -1, -1};
		private boolean completed = false;
		private boolean noException = true; 

		public void run() {

			// declaration
			long start, end;
			MatchingAndComputationTimes<DefaultEdge> matching;

			// convert
			try {
				file = Converter.convertFile(file);
			} catch (WrongFileFormatException wff) {
				System.out.print(". ");
				System.err.println(wff.getMessage());
				noException = false;
				return;
			}

			// create graph
			UndirectedGraph<String, DefaultEdge> graph = FileToGraph.fileToGraph(file);				

			// save number of nodes and edges
			vertexNumber = graph.vertexSet().size();
			edgeNumber = graph.edgeSet().size();

			// matching

			// compute matching and measure calculation time
			start = System.currentTimeMillis();
			matching = MatchingMaker.computeMatching(graph, algo, method);
			end = System.currentTimeMillis();
			// save size of matching
			size = (int)matching.getMatching().getWeight();
			// save calculation times
			time[0] = (end - start) / 1000.0;
			time[1] = matching.getTimeForDRR();
			time[2] = matching.getTimeForMatching();
			time[3] = matching.getTimeForPostprocessing();


			completed = true;
		}

		/**
		 * @return number of vertices
		 */
		public int getVertexNumber() {
			return vertexNumber;
		}

		/**
		 * @return number of edges
		 */
		public int getEdgeNumber() {
			return edgeNumber;
		}

		/**
		 * @return size of matching
		 */
		public int getSize() {
			return size;
		}

		/**
		 * @return calculation times
		 */
		public double[] getTime() {
			return time;
		}

		/**
		 * @return the completed
		 */
		public boolean isCompleted() {
			return completed;
		}

		/**
		 * @return the noException
		 */
		public boolean hasNoException() {
			return noException;
		}

		/**
		 * @return the file
		 */
		public File getFile() {
			return file;
		}

		/**
		 * @param file the file to set
		 */
		public void setFile(File file) {
			this.file = file;
		}

		/**
		 * @param method the kernelization method to set
		 */
		public void setKM(int method) {
			this.method = method;
		}

	}

}
