package kernelization;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Queue;
import java.util.Set;

import org.jgrapht.Graphs;
import org.jgrapht.UndirectedGraph;

/**
 * The class <code> Kernelization </code> makes kernelization for matching.
 * @author Viatcheslav Korenwein
 * @version 1.1
 */
public class Kernelization {
	
	/**
	 * Makes kernelization according to reduction rules 2.1 and 2.3.
	 * Graph will be changed.
	 * @param graph
	 * @return edges that have to be added to matching
	 */
	public static <V, E> Set<EdgeStorageCell<E>> kernelizeWithDegreeRestrictionUsedSpecificMethod(UndirectedGraph<V, E> graph){
		return kernelize(graph, 3);
	}

	/**
	 * Makes kernelization according to reduction rules 2.1 and 2.3. For rule 2.3 checks only degrees of neighbors and uses rule 2.2.
	 * Graph will be changed.
	 * @param graph
	 * @return edges that have to be added to matching
	 */ 
	public static <V, E> Set<EdgeStorageCell<E>> kernelizeWithDegreeRestrictionUsedMethodWithoudDegreeRestriction(UndirectedGraph<V, E> graph){
		return kernelize(graph, 2);
	}
		
	/**
	 * Makes kernelization according to reduction rules 2.1 and 2.2.
	 * Graph will be changed.
	 * @param graph
	 * @return edges that have to be added to matching
	 */
	public static <V, E> Set<EdgeStorageCell<E>> kernelizeWithoutDegreeRestriction(UndirectedGraph<V, E> graph){
		return kernelize(graph, 1);
	}
	
	/**
	 * Makes kernelization for matching.
	 * Graph will be changed.
	 * @param graph
	 * @param method - 1 for DRR without degree restriction,
	 * 2 for DRR with degree restriction used method for DRR without degree restriction,
	 * 3 for DRR with degree restriction used special method for it
	 * @return edges that have to be added to matching
	 */
	public static <V, E> Set<EdgeStorageCell<E>> kernelize(UndirectedGraph<V, E> graph, int method){
		// declaration
		Set<EdgeStorageCell<E>> edgesToMatching = new HashSet<EdgeStorageCell<E>>();
		
		// collect vertices with degree 0, 1 and 2
		Queue<V> verticesOfDeg0_1_2 = new LinkedList<V>();
		for (V vertex : graph.vertexSet()) {
			int deg = graph.degreeOf(vertex);
			if (deg == 0 || deg == 1 || deg == 2)
				verticesOfDeg0_1_2.add(vertex);
		}
		
		// data reduction rules
		while(! verticesOfDeg0_1_2.isEmpty()){
			// take first vertex from queue
			V vertex = verticesOfDeg0_1_2.poll();
			if (graph.containsVertex(vertex)){
				int deg = graph.degreeOf(vertex);
				
				if (deg == 0)
					// delete vertex
					graph.removeVertex(vertex);				
				
				if (deg == 1){
					// delete vertex and its neighbor and store the edge between them
					V neighbor = Graphs.neighborListOf(graph, vertex).get(0);
					
					if (method == 1) {
						Set<E> edgesOfNeighbor = graph.edgesOf(neighbor);
						// if neighbor of neighbor has degree = 3 -> add it to queue
						// otherwise if less then it was in queue, if greater then still not need
						for (E e : edgesOfNeighbor){
							V neighborOfNeighbor = Graphs.getOppositeVertex(graph, e, neighbor);
							if (graph.degreeOf(neighborOfNeighbor) == 3)
								verticesOfDeg0_1_2.add(neighborOfNeighbor);
						}
					}
					else {
						// neighbors of neighbor have to be checked whether their degree is 2 or 3
						// Possible, that nn was checked but its neighbor had degree >2.
						List<V> neighborsOfNeighbor = Graphs.neighborListOf(graph, neighbor);
						ListIterator<V> iter = neighborsOfNeighbor.listIterator();
						while(iter.hasNext()){
							V nn = iter.next();
							int deg_nn = graph.degreeOf(nn); 
							if (deg_nn == 2 || deg_nn == 3)
								verticesOfDeg0_1_2.add(nn);
						}
					}
					
					// store edge
					edgesToMatching.add(new EdgeStorageCell<>(graph.getEdge(vertex, neighbor)));
					// delete vertex and its neighbor
					graph.removeVertex(neighbor);
					graph.removeVertex(vertex);
				}
				
				if (deg == 2){
					// neighbors u and w
					List<V> neighbors = Graphs.neighborListOf(graph, vertex);
					V u = neighbors.get(0);
					V w = neighbors.get(1);
					int degU = graph.degreeOf(u);
					int degW = graph.degreeOf(w);
					
					if ( (method == 1) || (method == 2 && degU <= 2 && degW <= 2) ) {
						// merge neighbors of vertex
						V toDelete = merge(graph, u, w, verticesOfDeg0_1_2);
						
						// store both edges
						edgesToMatching.add(new EdgeStorageCell<>(graph.getEdge(vertex, u),graph.getEdge(vertex, w)));
						// remove vertex and one neighbor
						graph.removeVertex(vertex);
						graph.removeVertex(toDelete);
						
						if (method == 1) {						 
							// if merged vertex has degree < 3 and was not in queue -> add it to queue
							if (toDelete.equals(w) && degU > 2 && graph.degreeOf(u) <= 2)
								verticesOfDeg0_1_2.add(u);
							if (toDelete.equals(u) && degW > 2 && graph.degreeOf(w) <= 2)
								verticesOfDeg0_1_2.add(w);
						}
						else {		 
							// if merged vertex has degree < 3, add it to queue. Possible, that the merged vertex was checked but its neighbor had degree >2.
							if (toDelete.equals(w) && graph.degreeOf(u) <= 2)
								verticesOfDeg0_1_2.add(u);
							if (toDelete.equals(u) && graph.degreeOf(w) <= 2)
								verticesOfDeg0_1_2.add(w);
						}
					}

					if (method == 3 && degU <= 2 && degW <= 2){
						// delete vertex, merge u and w and store an edge
						E vu = graph.getEdge(vertex, u);
						E vw = graph.getEdge(vertex, w);
						graph.removeVertex(vertex);
						// if (u and w have the same neighbor) or (u and w are neighbors)
						// then we have only to remove w in order to merge u and w 
						if (degU > 1 && degW > 1){
							V neighborU = Graphs.neighborListOf(graph, u).get(0);
							V neighborW = Graphs.neighborListOf(graph, w).get(0);
							if (neighborU.equals(w)) {
								graph.removeVertex(w);
								// store edge from vertex to w
								edgesToMatching.add(new EdgeStorageCell<>(vw));
								// add u to queue. Possible, that u was checked but its neighbor had degree >2. Now u has degree 0.
								verticesOfDeg0_1_2.add(u);
							}
							else if (neighborU.equals(neighborW)) {
								graph.removeVertex(w);
								// store edge from vertex to w
								edgesToMatching.add(new EdgeStorageCell<>(vw));
								// check if neighbor of w has to be added to queue
								if (graph.degreeOf(neighborU) == 2)
									verticesOfDeg0_1_2.add(neighborU);
								// add u to queue. Possible, that u was checked but its neighbor had degree >2. Now u has degree 1.
								verticesOfDeg0_1_2.add(u);
							}
							// u and w have different neighbors
							else{
								// merge
								graph.addEdge(u, neighborW);
								graph.removeVertex(w);
								// store edge
								edgesToMatching.add(new EdgeStorageCell<E>(vw));
								// add u to queue. Possible, that u was checked but its neighbor had degree >2. Now u has degree 2, but has an other neighbor.
								verticesOfDeg0_1_2.add(u);
							}
						}
						// u has a neighbor and w hasn't
						else if (degU > 1){ 	//and degW = 1
							graph.removeVertex(w);
							edgesToMatching.add(new EdgeStorageCell<E>(vw));
							// add u to queue. Possible, that u was checked but its neighbor had degree >2. Now u has degree 1.
							verticesOfDeg0_1_2.add(u);
						}
						// w has a neighbor and u hasn't
						else if (degU == 1 && degW > 1){
							graph.removeVertex(u);
							edgesToMatching.add(new EdgeStorageCell<E>(vu));
							// add w to queue. Possible, that w had degree 2 and was checked but its neighbor had degree >2. Then w has degree 1 now.
							verticesOfDeg0_1_2.add(w);							
						}
						// u and w have no other neighbors except vertex
						else {
							graph.removeVertex(u);
							edgesToMatching.add(new EdgeStorageCell<E>(vu));	
						}
					}
				}
			}
		}
		return edgesToMatching;		
	}
	
	/**
	 * 
	 * Merges two vertices.
	 * @param graph
	 * @param u - the first vertex
	 * @param w - the second vertex
	 * @return vertex that has to be deleted
	 */
	public static <V, E> V merge(UndirectedGraph<V, E> graph, V u, V w, Queue<V> verticesOfDegLess3){
		int degU = graph.degreeOf(u);
		int degW = graph.degreeOf(w);
		if (degU >= degW){
			// merge u and w to u
			List<V> neighborsW = Graphs.neighborListOf(graph, w);
			for (V v : neighborsW){
				if (v != u)									// loops are not allowed
					// add edge. if already exists check if the other end vertex should be added to queue
					if (graph.addEdge(u, v) == null)
						if (graph.degreeOf(v) == 3)
							verticesOfDegLess3.add(v);
			}
			return w;
		}
		else{
			// merge u and w to w
			List<V> neighborsU = Graphs.neighborListOf(graph, u);
			for (V v : neighborsU){
				if (v != w)		//loops are not allowed
					// add edge. if already exists check if the other end vertex should be added to queue
					if (graph.addEdge(w, v) == null)
						if (graph.degreeOf(v) == 3)
							verticesOfDegLess3.add(v);
			}
			return u;
		}
			
	}
}
