package kernelization;
import java.util.Set;

import kernelization.EdgeStorageCell;
import kernelization.ToMatching;

import org.jgrapht.Graph;
import org.jgrapht.UndirectedGraph;
import org.jgrapht.alg.interfaces.MatchingAlgorithm;
import org.jgrapht.alg.interfaces.MatchingAlgorithm.Matching;
import org.jgrapht.alg.matching.EdmondsBlossomShrinking;
import org.jgrapht.alg.matching.GreedyWeightedMatching;
import org.jgrapht.alg.matching.PathGrowingWeightedMatching;

import blossom5_JNI.Blossom5;

/**
 * The class <code> MatchingMaker </code> computes a matching for given kernelization method and matching algorithm.
 * @author Viatcheslav Korenwein
 * @version 2.0
 * @param <V>
 * @param <E>
 */
public class MatchingMaker {

	/**
	 * Computes a matching.
	 * @param <V>
	 * @param <E>
	 * @param graph
	 * @param algNr
	 * @param kernMethod
	 * @return matching
	 */
	public static <V, E> MatchingAndComputationTimes<E> computeMatching(UndirectedGraph<V, E> graph, int algNr, int kernMethod){
		
		long start, end;
		double timeForDRR, timeForMatching, timeForPostprocessing;
		Matching<E> matching;	

		// choice of matching algorithm
		MatchingAlgorithm<V, E> matchingAlg = chooseMatchingAlgorithm(graph, algNr);

		// matching
		if (kernMethod == 0){
			start = System.currentTimeMillis();
			matching = matchingAlg.computeMatching();
			end = System.currentTimeMillis();
			timeForMatching = (end - start) / 1000.0;
			return new MatchingAndComputationTimes<E>(matching, 0.0, timeForMatching, 0.0);
		}

		// kernelization + matching
		else{
			start = System.currentTimeMillis();
			Set<EdgeStorageCell<E>> toMatching = kernelize(graph, kernMethod);
			end = System.currentTimeMillis();
			timeForDRR = (end - start) / 1000.0;

			start = System.currentTimeMillis();
			Matching<E> matching_k = matchingAlg.computeMatching();
			end = System.currentTimeMillis();
			timeForMatching = (end - start) / 1000.0;

			start = System.currentTimeMillis();
			matching = ToMatching.addToMatching(graph, matching_k, toMatching);
			end = System.currentTimeMillis();
			timeForPostprocessing = (end - start) / 1000.0;
			return new MatchingAndComputationTimes<E>(matching, timeForDRR, timeForMatching, timeForPostprocessing);
		}
	}

	/**
	 * Makes kernelization for matching.
	 * Graph will be changed.
	 * @param graph
	 * @param method - 1 for DRR without degree restriction,
	 * 2 for DRR with degree restriction used method for DRR without degree restriction,
	 * 3 (and default) for DRR with degree restriction used special method for it
	 * @return edges that have to be added to matching
	 */
	private static <V, E> Set<EdgeStorageCell<E>> kernelize(UndirectedGraph<V, E> graph, int method) {
		switch (method){
		case 1:
			return Kernelization.kernelizeWithoutDegreeRestriction(graph);
		case 2:
			return Kernelization.kernelizeWithDegreeRestrictionUsedMethodWithoudDegreeRestriction(graph);
		default:
			return Kernelization.kernelizeWithDegreeRestrictionUsedSpecificMethod(graph);
		}
	}


	/**
	 * Chooses matching algorithm.
	 * @param graph
	 * @param algNr - 1 (and default) for Edmonds Blossom Shrinking algorithm that runs in O(n^4), 
	 * 2 for 1/2-approximation greedy algorithm that runs in O(n + m log n), 
	 * 3 for a linear time 1/2-approximation algorithm that runs in O(m)
	 * @return new instance of MatchingAlgorithm
	 */
	private static <V, E> MatchingAlgorithm<V, E> chooseMatchingAlgorithm(Graph<V, E> graph, int algNr) {
		switch (algNr) {
		//	case 1:
		//		return new EdmondsBlossomShrinking<V, E>(graph);
		case 2:
			return new GreedyWeightedMatching<V, E>(graph);
		case 3:
			return new PathGrowingWeightedMatching<V, E>(graph);
		case 5:
			return new Blossom5<V, E>(graph);
		default:
			return new EdmondsBlossomShrinking<V, E>(graph);
		}
	}

}