package kernelization;

import java.util.HashSet;
import java.util.Set;

import org.jgrapht.UndirectedGraph;
import org.jgrapht.alg.interfaces.MatchingAlgorithm.Matching;
import org.jgrapht.alg.interfaces.MatchingAlgorithm.MatchingImpl;

/**
 * The class <code> ToMatching </code> adds edges from Set of EdgeStorageCell to given matching.
 * If EdgeStoregeCell contains one edge it will be added.
 * If EdgeStoregeCell contains two edges it will be checked which of these edges will be added.
 * @author Viatcheslav Korenwein
 * @version 1.0
 */
public class ToMatching {
	
	/**
	 * Adds edges to matching.
	 * @param graph
	 * @param matching
	 * @param edgesToMatching
	 * @return matching
	 * @throws IllegalArgumentException
	 */
	public static <V, E> Matching<E> addToMatching(UndirectedGraph<V, E> graph, Matching<E> matching, Set<EdgeStorageCell<E>> edgesToMatching)
		throws IllegalArgumentException{
		
		// get matched vertices
		Set<E> matching_k = new HashSet<E>(matching.getEdges());
		HashSet<V> matchedVertices = getVerticesFromEdgeSet(graph, matching_k);
		// add edges
		for(EdgeStorageCell<E> pair : edgesToMatching){
			if (pair.getEdge1() == null)
				throw new IllegalArgumentException("Not valid EdgeStorageCell. Edge1 = null.");
			// if EdgeStorageCell contains only one edge, add this edge to the matching
			if (pair.getEdge2() == null)
				matching_k.add(pair.getEdge1());
			// if EdgeStorageCell contains two edges, then we need to check which of these edges can be added to the matching
			else{
				// identify vertices of edges
				E e1 = pair.getEdge1();
				E e2 = pair.getEdge2();
				V s1 = graph.getEdgeSource(e1);
				V t1 = graph.getEdgeTarget(e1);
				V s2 = graph.getEdgeSource(e2);
				V t2 = graph.getEdgeTarget(e2);

				// find common vertex and call u the other vertex of e1
				V u;
				if (s1.equals(s2)){
					u = t1;
				}
				else if (s1.equals(t2)){
					u = t1;
				}
				else if (t1.equals(s2)){
					u = s1;
				}
				else if (t1.equals(t2)){
					u = s1;
				}
				else
					throw new IllegalArgumentException("Not valid EdgeStorageCell. Edges have no common vertex.");
				
				// if u is not matched we can add e1=(common vertex, u) to the matching, otherwise we add e2
				if (matchedVertices.contains(u))
					matching_k.add(e2);
				else
					matching_k.add(e1);
			}
		}
		
		// return matching	
		return new MatchingImpl<E>(matching_k, matching_k.size());
		
	}
	
	/**
	 * Returns vertices of edges in the set.
	 * @param graph
	 * @param edges
	 * @return
	 */
	public static <V, E> HashSet<V> getVerticesFromEdgeSet(UndirectedGraph<V, E> graph, Set<E> edges){		
		HashSet<V> vertices = new HashSet<V>();
		for (E edge : edges) {
			vertices.add(graph.getEdgeSource(edge));
			vertices.add(graph.getEdgeTarget(edge));
		}		
		return vertices;
	}
}
